{{
 ASC+Gameduino "Wireframe" Demo
 adapted from C source by Alessandro De Luca

 Based on "Wireframe" demo for Gameduino
 Copyright (c) 2011 by James Bowman <jamesb@excamera.com>
}}

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

OBJ
  es : "GD_eliteships_h"
  wf : "GD_wireframe_h"
' fp : "Float32"
  fp : "F32"
  gd : "GD_ASM"

VAR
  long _rseed, _oneus, _ptick, _usecs

  long sn, sm         ' Ship number (0-NSHIPS), Max ship number
  long cphi           ' Current rotation angle (float)
  long flip, plotting

  long gx, gy, gz     ' declared as static with initial value (see prolog)

PUB Main
  _PROLOG
  gd_setup
  gd_loop

PRI _PROLOG
  fp.start
  '_rseed := cnt
  _ptick := cnt
  _oneus := clkfreq / 1_000_000
  _usecs := 0
  gx := 0.5773503               '0.57735026918962573
  gy := 0.5773503               '0.57735026918962573
  gz := 0.5773503               '0.57735026918962573
  es.fix_pointers '<- work around for PropTool missing the @@@ operator
  sm := es.get_ship_max

PRI BLACK
  return GD.RGB(0,0,0)

PRI WHITE
  return GD.RGB(255,255,255)

PRI SWAP(ptra, ptrb) | swap_temp
  swap_temp  := long[ptra]
  long[ptra] := long[ptrb]
  long[ptrb] := swap_temp

PRI REPDIB(color)
' replicate a 2-bit color across the whole byte.
  return (color << 6) | (color << 4) | (color << 2) | color

PRI micros | ct, dt
  ct := cnt
  dt := (ct - _ptick) / _oneus
  _ptick := ct
  _usecs += dt
  return _usecs

PRI random(lim) | rval
  rval := (?_rseed) & $7FFFFFFF
  return (rval // lim)

CON
'////////////////////////////////////////////////////////////////////////////////
'//                                  Plotter
'////////////////////////////////////////////////////////////////////////////////
PRI pc_waitready
  repeat while (GD.m_rd(GD#COMM+7))

PRI pc_erase | color
  if flip
    color := 1
  else
    color := 2
  plotting := 0
  GD.m_wr(GD#J1_RESET, 1)
  GD.m_wr(GD#COMM+7, 1)
  GD.m_wr(GD#COMM+8, REPDIB(color ^ 3))
  GD.microcode(wf.eraser_code, wf.eraser_size)

PRI pc_begin | i, x, y, img, pal
'// Draw 256 sprites left to right, top to bottom, all in 4-color
'// palette mode.  By doing them in column-wise order, the address
'// calculation in setpixel is made simpler.
'// First 64 use bits 0-1, next 64 use bits 2-4, etc.
'// This gives a 256 x 256 4-color bitmap.
  repeat i from 0 to 255
    x := 72 + 16 * ((i >> 4) & 15)
    y := 22 + 16 * (i & 15)
    img := i & 63               'image 0-63
    pal := 3 - (i >> 6)         'palettes bits in columns 3,2,1,0
    GD.sprite(i, x, y, img, $08 | (pal << 1), 0, 0)
  flip := 0
  plotting := 0
  PC_erase
  PC_show

PRI pc_show
  pc_waitready
  if flip
    GD.m_wr16(GD#PALETTE4A + 0, BLACK)
    GD.m_wr16(GD#PALETTE4A + 2, WHITE)
    GD.m_wr16(GD#PALETTE4A + 4, BLACK)
    GD.m_wr16(GD#PALETTE4A + 6, WHITE)
  else
    GD.m_wr16(GD#PALETTE4A + 0, BLACK)
    GD.m_wr16(GD#PALETTE4A + 2, BLACK)
    GD.m_wr16(GD#PALETTE4A + 4, WHITE)
    GD.m_wr16(GD#PALETTE4A + 6, WHITE)
  flip ^= 1
  pc_erase

PRI pc_line(x0, y0, x1, y1) | steep, deltax, deltay, ystep, color ', error
  deltax := ||(x1 - x0)
  deltay := ||(y1 - y0)
  if deltay > deltax
    steep := 1
  else
    steep := 0
  if steep
    SWAP(@x0, @y0)
    SWAP(@x1, @y1)
  if x0 > x1
    SWAP(@x0, @x1)
    SWAP(@y0, @y1)
  deltax :=    x1 - x0
  deltay := ||(y1 - y0)
  if y0 < y1    
    ystep := 1
  else
    ystep := -1
  pc_waitready
  ifnot plotting
    GD.microcode(wf.wireframe_code, wf.wireframe_size)
    plotting := 1
    if flip
      color := 1
    else
      color := 2
    GD.m_wr(GD#COMM+8, color << 6)
  GD._wstart(GD#COMM+0)
  GD.transfer(x0)
  GD.transfer(y0)
  GD.transfer(x1)
  GD.transfer(y1)
  GD.transfer(steep)
  GD.transfer(deltax)
  GD.transfer(deltay)
  GD.transfer(ystep)
  GD._end

CON
'////////////////////////////////////////////////////////////////////////////////
'//                                  3D Projection
'////////////////////////////////////////////////////////////////////////////////

  OFS_V0 = 0
  OFS_V1 = 1

  OFS_X = 0
  OFS_Y = 1
  OFS_Z = 2

  WD_PROJSIZE = 2 'bytes
  WD_PROJSHAM = 1 'shift amount

  PROJSPACE_OFS_X = 128
  PROJSPACE_OFS_Y = 128
  PROJSPACE_SCALE = 140.0

  ROTATION_RATE   = 0.03 '0.02
  MAX_FRM_AT_ZERO = 199  '431

VAR
  long  mat[9]                  'static float mat[9]
  byte  projected[40 * 2]       'up to 40 points (byte[0]=X, byte[1]=Y)

PRI rotation(phi) | s, xs, ys, zs, c, ic
'-- Taken from glRotate()
  s := fp.Sin(phi)
  c := fp.Cos(phi)

  xs := fp.FMul(gx, s)
  ys := fp.FMul(gy, s)
  zs := fp.FMul(gz, s)
  ic := fp.FSub(1.0, c)

  mat[0] := fp.FAdd(fp.FMul(fp.FMul(gx, gx), ic), c )
  mat[1] := fp.FSub(fp.FMul(fp.FMul(gx, gy), ic), zs)
  mat[2] := fp.FAdd(fp.FMul(fp.FMul(gx, gz), ic), ys)

  mat[3] := fp.FAdd(fp.FMul(fp.FMul(gy, gx), ic), zs)
  mat[4] := fp.FAdd(fp.FMul(fp.FMul(gy, gy), ic), c )
  mat[5] := fp.FSub(fp.FMul(fp.FMul(gy, gz), ic), xs)

  mat[6] := fp.FSub(fp.FMul(fp.FMul(gz, gx), ic), ys)
  mat[7] := fp.FAdd(fp.FMul(fp.FMul(gz, gy), ic), xs)
  mat[8] := fp.FAdd(fp.FMul(fp.FMul(gz, gz), ic), c )

PRI project(sptr, fdist) | xx, yy, zz, x, y, z, q, vptr, vend, pptr
  vptr := word[sptr][es#ES_VERTPTR]
  vend := word[sptr][es#ES_NVERTS ]
  vend := (vend + (vend << 1)) + vptr
  pptr := @projected
  repeat while (vptr < vend)
    x  := fp.FFloat(~byte[vptr][OFS_X])
    y  := fp.FFloat(~byte[vptr][OFS_Y])
    z  := fp.FFloat(~byte[vptr][OFS_Z])
    xx := fp.FAdd(fp.FAdd(fp.FMul(x, mat[0]), fp.FMul(y, mat[3])), fp.FMul(z, mat[6]))
    yy := fp.FAdd(fp.FAdd(fp.FMul(x, mat[1]), fp.FMul(y, mat[4])), fp.FMul(z, mat[7]))
    zz := fp.FAdd(fp.FAdd(fp.FMul(x, mat[2]), fp.FMul(y, mat[5])), fp.FMul(z, mat[8]))
    zz := fp.FAdd(zz, fdist)
    q  := fp.FDiv(PROJSPACE_SCALE, fp.FAdd(PROJSPACE_SCALE, zz))
    byte[pptr][OFS_X] := PROJSPACE_OFS_X + fp.FRound(fp.FMul(xx, q))
    byte[pptr][OFS_Y] := PROJSPACE_OFS_Y + fp.FRound(fp.FMul(yy, q))
    vptr += es#ES_VERTSIZE
    pptr += WD_PROJSIZE

PRI draw(sptr, fdist) | v0, v1, eptr, eend, pptr
'  _sprintf(@msg, string("%3d.%d mt   "), fp.FTrunc(fdist), fp.FRound(fp.FMul(fdist, 100.0)) // 100)
'  GD.putstr(36, 1, @msg)
  project(sptr, fdist)
  eptr := word[sptr][es#ES_EDGEPTR]
  eend := word[sptr][es#ES_NEDGES ]
  eend := (eend << 1) + eptr
  pptr := @projected
  repeat while (eptr < eend)
    v0 := pptr + (byte[eptr][OFS_V0] << WD_PROJSHAM)
    v1 := pptr + (byte[eptr][OFS_V1] << WD_PROJSHAM)
    pc_line(byte[v0][OFS_X], byte[v0][OFS_Y], byte[v1][OFS_X], byte[v1][OFS_Y])
    eptr += es#ES_EDGESIZE

PRI cycle(fdist) | t, fps10
'-- Draw one frame of ship
  rotation(cphi)
  cphi := fp.FAdd(cphi, ROTATION_RATE)
  draw(es.get_ship_ptr(sn), fdist)
  'GD.waitvblank ' uncomment this to sync to 72Hz frame rate
  pc_show
  if (++every == 4)
    every := 0
    t := micros
    fps10 := 40_000_000 / (t - tprev)
    _sprintf(@msg, string("%3d.%d fps  "), fps10 / 10, fps10 // 10)
    GD.putstr(38, 0, @msg)
    tprev := t
'  _sprintf(@msg, string("%3d.%d mt   "), fp.FTrunc(fdist), fp.FRound(fp.FMul(fdist, 100.0)) // 100)
'  GD.putstr(36, 1, @msg)

PUB gd_setup
  GD.begin
  GD.ascii
  GD.putstr(0, 0, string("Accelerated wireframe"))
  PC_begin

PUB gd_loop | d, pship, pname, nzero
  every := 0
  tprev := 0
  repeat
    pship := es.get_ship_ptr(sn)
    pname := pship + es#ES_SHIPNAME
    GD.putstr(0, 36, string("                                                  "))
    GD.putstr(25 - (strsize(pname) >> 1), 36, pname)
{
    'added to show number of vertex/edges
    _sprintf(@msg, string("%3d/%d V/E  "), {
    }  word[pship][es#ES_NVERTS], word[pship][es#ES_NEDGES])
    GD.putstr(38, 1, @msg)
}
    'attempt to compensate for complexity affecting
    ' the time that one ship stays at max zoom
    nzero := 240 - (word[pship][es#ES_NEDGES] << 2)

    'decrement distance from 990.0 to 0.0
    repeat d from 99 to 0
      cycle(fp.FMul(10.0, fp.FFloat(d)))
    'stay at dist=0.0 for a while
'   repeat d from 0 to MAX_FRM_AT_ZERO
    repeat d from 0 to nzero
      cycle(0.0)
    'increment distance from 0.0 to 990.0
    repeat d from 0 to 99
      cycle(fp.FMul(10.0, fp.FFloat(d)))
    sn := (sn + 1) // sm

VAR
  long  tprev
  byte  every
  byte  msg[30]

PRI _pdec(ptr, value) | i, tst
  i := 1_000
  tst := 0
  repeat 4
    if value => i
      byte[ptr++] := (value / i + "0")
      value //= i
      tst~~
    elseif tst or i == 1
      byte[ptr++] := "0"
    i /= 10
  return ptr

PRI _sprintf(ptr, fmt, a, b) | i
'fake sprintf - only used for fps
  ptr := _pdec(ptr, a)
  byte[ptr++] := byte[fmt][3]
  ptr := _pdec(ptr, b)
  repeat i from 6 to 10
    byte[ptr++] := byte[fmt][i]
  byte[ptr] := 0

CON